#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include "wish.h"
#include "tunnel.h"

/* Usar threads y devices: /dev/tunnel
 * open rw, siempre funciona
 * ioctl: para definir el largo del tunel y para resetearlo
 * write(fd, &id, prior): pide entrar al tunel en pos 0, con mi id y prior
 * read(fd, &id, prior) : pide avanzar en una posicion, con mi id y prior
 *                   retorna 1 cuando me muevo
 *                   retorna EOF cuando sali del tunel
 * close(fd);        desocupa el tunel
 */

int nplayers;
int inter_x, inter_y;

int tunel; /* fd al tunel */

#define GORDO 10

tun_enter1(int tunel, int item) {
    write(tunel, &item, 2); /* Quiero entrar al tunel desde 1 */
}

tun_enter2(int tunel, int item) {
    write(tunel, &item, 1); /* Quiero entrar al tunel desde 2 */
}

int tun_move(int tunel, int item, int prior) {
    if(read(tunel, &item, prior) != prior) {
	delete_w(item);
	return 0; /* EOF */
    }

    move_w(item, GORDO, 0); 
    return 1;
}
 

/* Deberia ser un auto con un parametro de "entry" */
int auto2(int y) {
    int item;
    int i;
    int x = GORDO;

    item = create_obj_w(x, y, GORDO, "oval", "red");
    tun_enter2(tunel, item);
    coords_w(item, inter_x, inter_y, GORDO);

    while(tun_move(tunel, item, 2) != 0)
			usleep(150000);
    return 0;
}

int auto1(int y) {
    int item;
    int i;
    int x = GORDO;

    item = create_obj_w(x, y, GORDO, "oval", "green");
    tun_enter1(tunel, item);
    coords_w(item, inter_x, inter_y, GORDO);

    while(tun_move(tunel, item, 1) != 0)
    			usleep(150000);
    return 0;
}


main(int argc, char *argv[]) {
    int i, alive;
    int itxt1=0, itxt2=0;
    char s1[128], s2[128];
    int nautos, tot_autos=0;
    pthread_t pid;
    int waiting1, waiting2;
    pthread_t *hijos;

    if(argc!=3)
      {
        fprintf(stderr, "Use: %s nautos largo\n", argv[0]);
        exit(1);
      }

    nautos=atoi(argv[1]);
    nplayers=atoi(argv[2]);

    hijos = (pthread_t *) malloc((nautos+1)*sizeof(pthread_t));
    if(hijos == NULL) {
	fprintf(stderr, "out of mem\n");
	exit(1);
    }
    tunel = open("/dev/tunnel", 2);
    if(tunel < 0) {
	fprintf(stderr, "no pude abrir tunel\n");
	exit(1);
    }
    if(ioctl(tunel, TUNNEL_IOCRESET, NULL) != 0) {
	fprintf(stderr, "no es un tunel!\n");
	exit(1);
    }
    if(ioctl(tunel, TUNNEL_IOCSSIZE, &nplayers) != 0) {
	fprintf(stderr, "No pude hacer un tunel de tamaño %d!\n", nplayers);
	exit(1);
    }
    nplayers = 0;
    if(ioctl(tunel, TUNNEL_IOCGSIZE, &nplayers) != 0) {
	fprintf(stderr, "No pude hacer obtener tamaño del tunel!\n");
	exit(1);
    }

    inter_x = 3*GORDO;
    inter_y = 3*GORDO;
    waiting1 = waiting2 = 0;

    init_w((nplayers+1)*GORDO+6*GORDO, 6*GORDO);

    create_obj_w(inter_x, inter_y, GORDO, "rectangle", "white");


   sleep(1); 

    while(tot_autos < nautos) {
        /* usamos 120 contra 32 para mostrar que funciona */
	if((rand()>>24) < 120)
	    if(pthread_create(&hijos[tot_autos++], NULL, auto1, (void *)(2*GORDO)) != 0) {
	tot_autos--;
	break;
          }
	if((rand()>>24) < 32)
	    if(pthread_create(&hijos[tot_autos++], NULL, auto2, (void *)(4*GORDO)) != 0) {
	tot_autos--;
	break;
	 }
	waiting1 = 2;
 	ioctl(tunel, TUNNEL_IOCWAIT, &waiting1);
	waiting2 = 1;
 	ioctl(tunel, TUNNEL_IOCWAIT, &waiting2);
        sprintf(s1, "cnt=%d", waiting1);
        sprintf(s2, "cnt=%d", waiting2);
    	if(itxt1 != 0) delete_w(itxt1);
    	if(itxt2 != 0) delete_w(itxt2);
        itxt1 = create_obj_w(3*GORDO, GORDO, GORDO, "text", s1);
        itxt2 = create_obj_w(3*GORDO, 5*GORDO+5, GORDO, "text", s2);
        usleep(140000);
    }

    for(i = 0; i < tot_autos; i++) {
	pthread_join(hijos[i], NULL);
	usleep(100000);
	waiting1 = 2;
 	ioctl(tunel, TUNNEL_IOCWAIT, &waiting1);
	waiting2 = 1;
 	ioctl(tunel, TUNNEL_IOCWAIT, &waiting2);
        sprintf(s1, "cnt=%d", waiting1);
        sprintf(s2, "cnt=%d", waiting2);
    	if(itxt1 != 0) delete_w(itxt1);
    	if(itxt2 != 0) delete_w(itxt2);
        itxt1 = create_obj_w(3*GORDO, GORDO, GORDO, "text", s1);
        itxt2 = create_obj_w(3*GORDO, 5*GORDO+5, GORDO, "text", s2);
    }
    close_w("Finish!", 2*GORDO, GORDO);
}
